import enum
from .pyast import transform_stmt_block, CodeBlockType, add_scope_context, \
    get_frame_local, AlwaysWrapper
from .util import clog2, max_value, cast, VarCastType
from .stmts import if_, switch_, IfStmt, SwitchStmt
from .ports import PortBundle
from .fsm import FSM
from .interface import InterfaceWrapper
import _kratos
from _kratos import get_fn_ln
from typing import List, Dict, Union, Tuple

__GLOBAL_DEBUG = False


def set_global_debug(value: bool):
    global __GLOBAL_DEBUG
    assert isinstance(value, bool)
    __GLOBAL_DEBUG = value


def get_global_debug():
    return __GLOBAL_DEBUG


# this is a wrapper python class to interface with the underlying python
# binding
class PortDirection(enum.Enum):
    In = _kratos.PortDirection.In
    Out = _kratos.PortDirection.Out
    InOut = _kratos.PortDirection.InOut


class PortType(enum.Enum):
    Data = _kratos.PortType.Data
    Clock = _kratos.PortType.Clock
    AsyncReset = _kratos.PortType.AsyncReset
    ClockEnable = _kratos.PortType.ClockEnable
    Reset = _kratos.PortType.Reset


class BlockEdgeType(enum.Enum):
    Posedge = _kratos.BlockEdgeType.Posedge
    Negedge = _kratos.BlockEdgeType.Negedge


class StatementBlockType(enum.Enum):
    Combinational = _kratos.StatementBlockType.Combinational
    Sequential = _kratos.StatementBlockType.Sequential
    Initial = _kratos.StatementBlockType.Initial
    Latch = _kratos.StatementBlockType.Latch


class CodeBlock:
    def __init__(self, generator, block_type: StatementBlockType,
                 debug_frame_depth):
        self.block_type = block_type
        self._generator = generator
        if block_type == StatementBlockType.Combinational:
            self._block = generator.internal_generator.combinational()
        elif block_type == StatementBlockType.Initial:
            self._block = generator.internal_generator.initial()
        elif block_type == StatementBlockType.Latch:
            self._block = generator.internal_generator.latch()
        else:
            self._block = generator.internal_generator.sequential()

        if generator.debug:
            fn, ln = get_fn_ln(debug_frame_depth)
            self._block.add_fn_ln((fn, ln))

    def add_stmt(self, stmt, add_fn_ln: bool = True, depth=2):
        if hasattr(stmt, "stmt"):
            self._block.add_stmt(stmt.stmt())
        else:
            self._block.add_stmt(stmt)
        if add_fn_ln and self._generator.debug:
            info = get_fn_ln(depth)
            stmt.add_fn_ln(info)
            add_scope_context(stmt, get_frame_local(depth))

    def remove_stmt(self, stmt):
        if hasattr(stmt, "stmt"):
            self._block.remove_stmt(stmt.stmt())
        else:
            self._block.remove_stmt(stmt)

    def stmt(self):
        return self._block

    def if_(self, predicate: _kratos.Var) -> IfStmt:
        stmt = if_(predicate)
        self.add_stmt(stmt.stmt(), depth=3)
        return stmt

    def switch_(self, predicate: _kratos.Var) -> SwitchStmt:
        stmt = switch_(predicate)
        self.add_stmt(stmt.stmt(), depth=3)
        return stmt

    def __getitem__(self, item):
        return self._block[item]

    def add_attribute(self, attr):
        self._block.add_attribute(attr)


class SequentialCodeBlock(CodeBlock):
    def __init__(self, generator, sensitivity_list,
                 debug_frame_depth: int = 4):
        super().__init__(generator, StatementBlockType.Sequential,
                         debug_frame_depth)
        for cond, var in sensitivity_list:
            assert isinstance(cond, BlockEdgeType)
            assert isinstance(var, _kratos.Var)
            self._block.add_condition([cond.value, var])


class CombinationalCodeBlock(CodeBlock):
    def __init__(self, generator,
                 debug_frame_depth: int = 4):
        super().__init__(generator, StatementBlockType.Combinational,
                         debug_frame_depth)


class InitialCodeBlock(CodeBlock):
    def __init__(self, generator,
                 debug_frame_depth: int = 4):
        super().__init__(generator, StatementBlockType.Initial,
                         debug_frame_depth)


class LatchCodeBlock(CodeBlock):
    def __init__(self, generator,
                 debug_frame_depth: int = 4):
        super().__init__(generator, StatementBlockType.Latch,
                         debug_frame_depth)


def enum(name, definition, width=None):
    if isinstance(name, _kratos.Var):
        assert isinstance(definition, _kratos.Enum)
        v = cast(name, VarCastType.Enum, enum_type=definition)
        return v
    if isinstance(definition, (list, tuple)):
        defs = {}
        for n in definition:
            if isinstance(n, (list, tuple)):
                defs[n[0]] = n[1]
            else:
                defs[n] = len(defs)
        definition = defs
    if width is None:
        width = clog2(max_value(definition) + 1)
    return Generator.get_context().enum(name, definition, width)


def has_enum(name):
    return Generator.get_context().has_enum(name)


class PortProxy:
    def __init__(self, generator: "Generator"):
        self.__generator = generator

    def __getitem__(self, key):
        if self.__generator.internal_generator.has_port_bundle(key):
            p = self.__generator.internal_generator.get_bundle_ref(key)
        else:
            p = self.__generator.internal_generator.get_port(key)
        if p is None:
            raise AttributeError("{0} doesn't exist".format(key))
        return p

    def __getattr__(self, key):
        return self[key]

    def __contains__(self, key):
        return self.__generator.internal_generator.has_port(key)

    def __iter__(self):
        return self.__generator.internal_generator.ports_iter()


class ParamProxy:
    def __init__(self, generator: "Generator"):
        self.__generator = generator

    def __getitem__(self, key):
        p = self.__generator.internal_generator.get_param(key)
        if p is None:
            raise AttributeError("{0} doesn't exist".format(key))
        return p

    def __getattr__(self, key):
        return self[key]

    def __contains__(self, item):
        return self.__generator.internal_generator.get_param(item) is not None

    def __iter__(self):
        return self.__generator.internal_generator.param_iter()


class VarProxy:
    def __init__(self, generator):
        self.__generator = generator.internal_generator

    def __getitem__(self, key):
        v = self.__generator.get_var(key)
        if v is None:
            raise AttributeError("{0} doesn't exist".format(key))
        return v

    def __getattr__(self, key):
        return self[key]

    def __contains__(self, key):
        return self.__generator.has_var(key)

    def __iter__(self):
        return self.__generator.vars_iter()


class InterfaceProxy:
    def __init__(self, generator):
        self.__generator = generator.internal_generator

    def __getitem__(self, item):
        return self.__generator.get_interface(item)

    def __getattr__(self, key):
        return self[key]


class GeneratorMeta(type):
    def __init__(cls, name, bases, attrs):
        super().__init__(name, bases, attrs)
        cls._cache = {}


class Generator(metaclass=GeneratorMeta):
    __context = _kratos.Context()
    __inspect_frame_depth: int = 2

    def __init__(self, name: str, debug: bool = False, is_clone: bool = False,
                 internal_generator=None):
        """
        Base class for all generators
        :param name: generator name
        :param debug: set to ``True`` if you want to collect debug information
        on this particular generator
        :param is_clone: mark whether the generator is a clone or not.
        :param internal_generator: native C++ handle
        """
        # for initialization
        self.__cached_initialization = []

        if internal_generator is not None:
            assert isinstance(internal_generator, _kratos.Generator)
            self.__generator = internal_generator
        else:
            if not is_clone and len(name) > 0:
                self.__generator = self.__context.generator(name)
            else:
                self.__generator = self.__context.empty_generator()
                self.__generator.is_cloned = True
            self.__set_generator_name(name)

        self.__child_generator: Dict[str, Generator] = {}

        if not debug:
            self.debug = get_global_debug()
        else:
            self.debug = debug

        if debug:
            fn, ln = get_fn_ln(Generator.__inspect_frame_depth)
            self.__generator.add_fn_ln((fn, ln))

        # gemstone style port interface
        self.ports = PortProxy(self)
        self.params = ParamProxy(self)
        self.vars = VarProxy(self)
        self.interfaces = InterfaceProxy(self)

        self.__def_instance = self

        # helper function data
        self.__reg_next_stmt = {}
        self.__reg_init_stmt = {}
        self.__reg_en_stmt = {}

        # meta data
        self.__stmt_label_mapping = {}
        self.__parent = None

    def __getitem__(self, instance_name: str):
        """
        Get child instance through instance name
        :param instance_name: instance name of the child generator
        :return: Child generator
        """
        assert instance_name in self.__child_generator, \
            "{0} does not exist in {1}".format(instance_name,
                                               self.instance_name)
        return self.__child_generator[instance_name]

    @property
    def def_instance(self):
        """
        The definition instance of this generator. It can be itself or
        the clone reference
        :return: definition instance
        """
        return self.__def_instance

    @property
    def name(self):
        """
        Generator name usually corresponds to the name of the module. However,
        if unification happens, its verilog name will change.
        :return: The name of the generator
        """
        return self.__generator.name

    @name.setter
    def name(self, name: str):
        self.__generator.name = name

    @property
    def instance_name(self):
        """
        Instance name of a generator. It has to be unique within a parent
        generator.
        :return: the instance name of the generator
        """
        return self.__generator.instance_name

    @instance_name.setter
    def instance_name(self, name: str):
        if self.__parent is not None:
            old_name = self.__generator.instance_name
            ref = self.__parent.__child_generator.pop(old_name)
            assert ref == self
            self.__parent.__child_generator[name] = self
        self.__generator.instance_name = name

    @property
    def is_stub(self):
        """
        If a generator is mark as a stub, most of the passes won't touch it
        and it's the user's responsibility to keep track of it. Kratos will
        attempt to zero out the stub outputs.
        :return: ``True`` if it's a stub
        """
        return self.__generator.is_stub()

    @is_stub.setter
    def is_stub(self, value: bool):
        self.__generator.set_is_stub(value)

    @property
    def external(self):
        """
        External module typically is used when importing external verilog files.
        If set from user, all the passes and code gen will skip this generator
        definition.
        :return: ``True`` if it's an external generator
        """
        return self.__generator.external()

    @external.setter
    def external(self, value: bool):
        self.__generator.set_external(value)

    @property
    def debug(self):
        return self.__generator.debug

    @debug.setter
    def debug(self, value):
        self.__generator.debug = value

    @property
    def is_cloned(self):
        return self.__generator.is_cloned

    @property
    def stmts_count(self):
        return self.__generator.stmts_count()

    def get_stmt_by_index(self, index):
        return self.__generator.get_stmt(index)

    def var(self, name: str, width: Union[int, _kratos.Param, _kratos.Enum,
                                          _kratos.PackedStruct],
            is_signed: bool = False, size: Union[int, Union[List, Tuple]] = 1,
            packed: bool = False, explicit_array: bool = False) -> _kratos.Var:
        size, params = self.__filter_size(size)
        if isinstance(width, _kratos.Enum):
            v = self.__generator.enum_var(name, width)
        elif isinstance(width, _kratos.PackedStruct):
            v = self.__generator.var_packed(name, width, size)
        else:
            v = self.__generator.var(name, width, size, is_signed)
        if self.debug:
            v.add_fn_ln(get_fn_ln())
        if not isinstance(width, _kratos.PackedStruct):
            v.is_packed = packed
            v.explicit_array = explicit_array
        self.__set_var_size(v, params)
        return v

    def combinational(self):
        if self.is_cloned:
            self.__cached_initialization.append((self.combinational, []))
            return
        return CombinationalCodeBlock(self, 3)

    def sequential(self, *sensitivity_list: Tuple[BlockEdgeType,
                                                  _kratos.Var]):
        if self.is_cloned:
            self.__cached_initialization.append((self.sequential,
                                                 [sensitivity_list]))
            return
        return SequentialCodeBlock(self, sensitivity_list, 3)

    @staticmethod
    def __filter_size(size):
        if isinstance(size, int):
            return size, {}
        else:
            if isinstance(size, _kratos.Var):
                return 1, {0: size}
            assert isinstance(size, (list, tuple))
            result = []
            params = {}
            for i in range(len(size)):
                var = size[i]
                if isinstance(var, _kratos.Var):
                    params[i] = var
                    var = 1
                result.append(var)
            return result, params

    @staticmethod
    def __set_var_size(var, params: Dict[int, _kratos.Var]):
        for index, param in params.items():
            var.set_size_param(index, param)

    def port(self, name: str, width: Union[int, _kratos.Param, _kratos.Enum],
             direction: PortDirection,
             port_type: PortType = PortType.Data,
             is_signed: bool = False, size: Union[int, Union[List, Tuple]] = 1,
             packed: bool = False,
             explicit_array: bool = False) -> _kratos.Port:
        size, params = self.__filter_size(size)
        if isinstance(width, _kratos.Enum):
            p = self.__generator.port(direction.value, name, width)
        else:
            p = self.__generator.port(direction.value, name, width, size,
                                      port_type.value, is_signed)
        if self.debug:
            p.add_fn_ln(get_fn_ln())
        p.is_packed = packed
        p.explicit_array = explicit_array
        self.__set_var_size(p, params)
        return p

    def port_from_def(self, port: _kratos.Port, name="", check_param: bool = True):
        if name:
            return self.__generator.port(port, name, check_param)
        else:
            return self.__generator.port(port, check_param)

    def var_from_def(self, var: _kratos.Var, name):
        return self.__generator.var(var, name)

    def param_from_def(self, param: _kratos.Param, name=None):
        return self.__generator.parameter(param, name)

    def input(self, name, width: Union[int, _kratos.Param, _kratos.Enum,
                                       _kratos.PackedStruct],
              port_type: PortType = PortType.Data,
              is_signed: bool = False, size: Union[int, Union[List, Tuple]] = 1,
              packed: bool = False,
              explicit_array: bool = False) -> _kratos.Port:
        size, params = self.__filter_size(size)
        if isinstance(width, _kratos.Enum):
            p = self.__generator.port(PortDirection.In.value, name, width)
        elif isinstance(width, _kratos.PackedStruct):
            p = self.__generator.port_packed(PortDirection.In.value, name,
                                             width, size)
        else:
            p = self.__generator.port(PortDirection.In.value, name, width, size,
                                      port_type.value, is_signed)
        if self.debug:
            p.add_fn_ln(get_fn_ln())
        if not isinstance(width, _kratos.PackedStruct):
            p.is_packed = packed
            p.explicit_array = explicit_array
        self.__set_var_size(p, params)
        return p

    def clock(self, name, is_input=True):
        direction = PortDirection.In if is_input else PortDirection.Out
        p = self.__generator.port(direction.value, name, 1, 1,
                                  PortType.Clock.value, False)
        if self.debug:
            p.add_fn_ln(get_fn_ln())
        return p

    def clock_en(self, name, is_input=True):
        direction = PortDirection.In if is_input else PortDirection.Out
        p = self.__generator.port(direction.value, name, 1, 1,
                                  PortType.ClockEnable.value, False)
        if self.debug:
            p.add_fn_ln(get_fn_ln())
        return p

    def reset(self, name, is_input=True, is_async=True, active_high=None):
        direction = PortDirection.In if is_input else PortDirection.Out
        reset = PortType.AsyncReset if is_async else PortType.Reset
        p = self.__generator.port(direction.value, name, 1, 1, reset.value,
                                  False)
        if self.debug:
            p.add_fn_ln(get_fn_ln())
        if active_high is not None:
            p.active_high = active_high
        return p

    def output(self, name, width: Union[int, _kratos.Param, _kratos.Enum,
                                        _kratos.PackedStruct],
               port_type: PortType = PortType.Data,
               is_signed: bool = False,
               size: Union[int, Union[List, Tuple]] = 1, packed: bool = False,
               explicit_array: bool = False) -> _kratos.Port:
        size, params = self.__filter_size(size)
        if isinstance(width, _kratos.Enum):
            p = self.__generator.port(PortDirection.Out.value, name, width)
        elif isinstance(width, _kratos.PackedStruct):
            p = self.__generator.port_packed(PortDirection.Out.value, name,
                                             width, size)
        else:
            p = self.__generator.port(PortDirection.Out.value, name, width,
                                      size,
                                      port_type.value, is_signed)
        if self.debug:
            p.add_fn_ln(get_fn_ln())
        if not isinstance(width, _kratos.PackedStruct):
            p.is_packed = packed
            p.explicit_array = explicit_array
        self.__set_var_size(p, params)
        return p

    def var_packed(self, name: str, struct_packed: _kratos.PortPackedStruct):
        v = self.__generator.var_packed(name, struct_packed)
        if self.debug:
            v.add_fn_ln(get_fn_ln())
        return v

    def port_bundle(self, bundle_name, bundle: PortBundle):
        assert isinstance(bundle, PortBundle)
        if self.debug:
            return self.__generator.add_bundle_port_def(bundle_name,
                                                        bundle.definition,
                                                        get_fn_ln())
        else:
            return self.__generator.add_bundle_port_def(bundle_name,
                                                        bundle.definition)

    def parameter(self, name: str, width: int = 32, value=None,
                  is_signed: bool = False, initial_value=None,
                  is_raw_type: bool = False) -> _kratos.Param:
        if is_raw_type:
            param = self.__generator.parameter(name)
        else:
            if isinstance(width, _kratos.Enum):
                param = self.__generator.parameter(name, width)
            else:
                param = self.__generator.parameter(name, width, is_signed)
        if value is not None:
            param.value = value
        if initial_value is not None:
            param.initial_value = initial_value
            # set value as well
            if value is None:
                param.value = initial_value
        if self.debug:
            fn, ln = get_fn_ln()
            param.add_fn_ln((fn, ln))
        return param

    # alias
    param = parameter

    def enum(self, name: str, values: Dict[str, int], width=None):
        if width is None:
            width = clog2(max_value(values) + 1)
        return self.__generator.enum(name, values, width)

    def enum_var(self, name: str, def_: _kratos.Enum):
        return self.__generator.enum_var(name, def_)

    def interface(self, interface, name, is_port: bool = False):
        return InterfaceWrapper(
            self.__generator.interface(interface, name, is_port))

    def get_var(self, name):
        return self.__generator.get_var(name)

    def remove_port(self, port_name):
        assert self.__generator.has_port(port_name)
        self.__generator.remove_port(port_name)

    def remove_var(self, var_name):
        assert self.__generator.has_var(var_name)
        self.__generator.remove_var(var_name)

    def add_attribute(self, attr):
        self.__generator.add_attribute(attr)

    def find_attribute(self, func):
        return self.__generator.find_attribute(func)

    @property
    def internal_generator(self):
        return self.__generator

    def add_always(self, fn, comment="", label="", sensitivity=None,
                   fn_ln=None, unroll_for=False, ssa_transform=False,
                   **kargs):
        if self.is_cloned:
            # TODO: fix this with kargs
            self.__cached_initialization.append((self.add_code, [fn, comment]))
            return
        block_type, raw_sensitives, stmts = transform_stmt_block(self, fn, unroll_for=unroll_for,
                                                                 fn_ln=fn_ln, kargs=kargs,
                                                                 apply_ssa=ssa_transform)
        if sensitivity:
            # override the block type and sensivitives
            block_type = CodeBlockType.Sequential
            raw_sensitives = sensitivity
        if block_type == CodeBlockType.Combinational:
            # it's a combinational block
            comb = CombinationalCodeBlock(self)
            for stmt in stmts:
                comb.add_stmt(stmt, False)
            node = comb
        elif block_type == CodeBlockType.Initial:
            # it's a initial block
            init = InitialCodeBlock(self)
            for stmt in stmts:
                init.add_stmt(stmt, False)
            node = init
        elif block_type == CodeBlockType.Latch:
            # it's a latch block
            latch = LatchCodeBlock(self)
            for stmt in stmts:
                latch.add_stmt(stmt, False)
            node = latch
        else:
            sensitivity_list = []
            for edge, var_name in raw_sensitives:
                edge = BlockEdgeType[edge]
                if isinstance(var_name, str):
                    var = self.internal_generator.get_var(var_name)
                else:
                    assert isinstance(var_name, _kratos.Var)
                    var = var_name
                sensitivity_list.append((edge, var))
            seq = SequentialCodeBlock(self, sensitivity_list)
            for stmt in stmts:
                seq.add_stmt(stmt, False)
            node = seq
        # add context vars
        if self.debug and fn_ln is None:
            for stmt in stmts:
                add_scope_context(stmt, get_frame_local(2))
        if comment:
            node.comment = comment
        if label:
            self.mark_stmt(label, node)
        if len(kargs) > 0:
            from .passes import Attribute
            for key, value in kargs.items():
                bool_val = bool(value)
                if bool_val:
                    node.add_attribute(Attribute.create(key))
        if ssa_transform:
            from .passes import Attribute
            node.add_attribute(Attribute.create("ssa"))
        return node

    add_code = add_always

    def __assign(self, var_to, var_from):
        correct_dir, correct_assign = self.__generator.correct_wire_direction(
            var_to, var_from)
        if not correct_assign:
            raise ValueError(str(var_to) + " cannot be assign to " +
                             str(var_from) +
                             ". Please check your module hierarchy")
        if correct_dir:
            stmt = var_to.assign(var_from)
        else:
            stmt = var_from.assign(var_to)
        self.add_stmt(stmt, False)
        return stmt

    def wire(self, var_to, var_from,
             attributes: Union[List[_kratos.passes.Attribute],
                               _kratos.passes.Attribute] = None,
             comment="", locals_=None, fn_ln=None, additional_frame=0,
             no_fn_ln=False):
        if self.is_cloned:
            if self.debug and locals_ is None:
                locals_ = get_frame_local(2 + additional_frame)
            self.__cached_initialization.append((self.wire, [var_to, var_from,
                                                             attributes,
                                                             comment, locals_]))
            return
        # wire interface is a special treatment
        if isinstance(var_from, (_kratos.InterfaceRef, InterfaceWrapper)) or \
                isinstance(var_to, (_kratos.InterfaceRef, InterfaceWrapper)):
            if isinstance(var_from, InterfaceWrapper):
                var_from = var_from.internal_interface
            if isinstance(var_to, InterfaceWrapper):
                var_to = var_to.internal_interface
            assert isinstance(var_from, _kratos.InterfaceRef)
            assert isinstance(var_to, _kratos.InterfaceRef)
            # TODO: add debug info to interface wiring
            self.__generator.wire_interface(var_to, var_from)
            return
        # this is a top level direct wire assignment
        # notice that we can figure out the direction automatically if
        # both of them are ports
        # handle port bundles
        if isinstance(var_to, _kratos.PortBundleRef):
            assert isinstance(var_from, _kratos.PortBundleRef)
            if self.debug:
                entry = get_fn_ln(2 + additional_frame)
            else:
                entry = []
            var_from.assign(var_to, self.__generator, entry)
            return
        if isinstance(var_to, _kratos.Port) and isinstance(var_from,
                                                           _kratos.Port):
            stmt = self.__generator.wire_ports(var_to, var_from)
        else:
            stmt = self.__assign(var_to, var_from)

        if self.debug and not no_fn_ln:
            if fn_ln is not None:
                stmt.add_fn_ln(fn_ln)
            else:
                stmt.add_fn_ln(get_fn_ln(2 + additional_frame))
                if locals_ is None:
                    locals_ = get_frame_local(2 + additional_frame)
                add_scope_context(stmt, locals_)

        if attributes is not None:
            if not isinstance(attributes, list):
                attributes = [attributes]
            for attr in attributes:
                stmt.add_attribute(attr)

        if comment:
            stmt.comment = comment

    def add_fsm(self, fsm_name: str, clk_name=None, reset_name=None,
                reset_high=True):
        if clk_name is not None and reset_name is not None:
            clk = self.__generator.get_var(clk_name)
            reset = self.__generator.get_var(reset_name)
            fsm = FSM(self, self.__generator.fsm(fsm_name, clk, reset))
        else:
            fsm = FSM(self, self.__generator.fsm(fsm_name))
        fsm.set_reset_high(reset_high)
        return fsm

    def property(self, property_name: str, seq):
        return self.__generator.property(property_name, seq)

    def add_stmt(self, stmt, add_ln_info=True):
        if self.is_cloned:
            self.__cached_initialization.append((self.add_stmt, [stmt]))
            return
        self.__generator.add_stmt(stmt)
        if add_ln_info and self.debug:
            stmt.add_fn_ln(get_fn_ln())
            add_scope_context(stmt, get_frame_local(2))

    def remove_stmt(self, stmt):
        if self.is_cloned:
            self.__cached_initialization.append((self.remove_stmt, [stmt]))
            return
        self.__generator.remove_stmt(stmt)

    def add_child_generator(self, instance_name: str, generator: "Generator",
                            comment="", python_only=False, **kargs):
        if self.is_cloned:
            self.__cached_initialization.append((self.add_child_generator,
                                                 (instance_name, generator)))
            return
        if instance_name in self.__child_generator:
            raise Exception(
                "{0} already exists in {1}".format(instance_name,
                                                   self.instance_name))
        assert isinstance(generator,
                          Generator), "generator is not a Generator instance"

        self.__child_generator[instance_name] = generator
        generator.__parent = self

        if python_only:
            # only add it to the python level interface, the caller is
            # responsible for the connections etc
            return
        else:
            generator.__generator.instance_name = instance_name

        if self.debug:
            fn, ln = get_fn_ln()
            self.__generator.add_child_generator(instance_name,
                                                 generator.__generator,
                                                 (fn, ln))
        else:
            self.__generator.add_child_generator(instance_name,
                                                 generator.__generator)
        if comment:
            self.__generator.set_child_comment(instance_name, comment)

        # set parameter values first
        for child_param, parent_value in kargs.items():
            if child_param in generator.params:
                generator.params[child_param].value = parent_value

        # bulk wiring
        for child_port, parent_port in kargs.items():
            # it can be parameter as well. we already taken care of that
            if child_port in generator.params:
                continue
            if isinstance(parent_port, str):
                if parent_port in self.ports:
                    parent_port = self.ports[parent_port]
                else:
                    parent_port = self.vars[parent_port]
            self.wire(generator.ports[child_port], parent_port,
                      additional_frame=1)

    # alias
    add_child = add_child_generator

    def remove_child_generator(self, generator):
        if self.is_cloned:
            self.__cached_initialization.append((self.remove_child_generator,
                                                 [generator]))
            return
        if generator.instance_name not in self.__child_generator:
            raise Exception("{0} doesn't exist in {1}".format(generator.name,
                                                              self.name))
        self.__child_generator.pop(generator.instance_name)
        self.__generator.remove_child_generator(generator.__generator)

    def replace(self, child_name: str, new_child: "Generator"):
        assert child_name in self.__child_generator
        if self.debug:
            debug_info = get_fn_ln()
            self.__generator.replace(child_name, new_child.internal_generator,
                                     debug_info)
        else:
            self.__generator.replace(child_name, new_child.internal_generator)
        self.__child_generator[child_name] = new_child

    def child_generator(self):
        return self.__child_generator

    @staticmethod
    def clear_context():
        Generator.__context.clear()
        # also clean the caches
        clses = Generator.__subclasses__()
        for cls in clses:  # type: Generator
            cls._cache.clear()
        # clean the function calls
        from .func import clear_context
        clear_context()

    @staticmethod
    def clear_context_hash():
        Generator.__context.clear_hash()

    @staticmethod
    def get_context():
        return Generator.__context

    @staticmethod
    def from_verilog(top_name: str, src_file: str, lib_files: List[str],
                     port_mapping: Dict[str, PortType]):
        g = Generator("")
        _port_mapping = {}
        for name, _type in port_mapping.items():
            _port_mapping[name] = _type.value
        g.__generator = _kratos.Generator.from_verilog(Generator.__context,
                                                       src_file, top_name,
                                                       lib_files, _port_mapping)
        Generator.__context.add(g.__generator)
        return g

    def __contains__(self, generator: "Generator"):
        if not isinstance(generator, (Generator, _kratos.Generator)):
            return False
        elif isinstance(generator, Generator):
            return generator.__generator in self.__generator
        else:
            return generator in self.__generator

    def initialize_clone(self):
        if self.is_cloned:
            self.__generator.is_cloned = False
            for fn, args in self.__cached_initialization:
                fn(*args)
            self.__cached_initialization.clear()

    def __set_generator_name(self, name):
        self.__generator.name = name

    @classmethod
    def __cached_py_generator(cls, **kargs):
        hash_value = 0
        for key, value in kargs.items():
            hash_value = hash_value ^ (hash(key) << 16) ^ hash(value)
        if hash_value not in cls._cache:
            g = cls(**kargs)
            cls._cache[hash_value] = g
            return g, False
        else:
            return cls._cache[hash_value], True

    @classmethod
    def clone(cls, **kargs):
        gen, cached = cls.__cached_py_generator(**kargs)
        if not cached:
            return gen
        else:
            g = Generator("")
            g.__generator = gen.__generator.clone()
            g.__def_instance = gen
            return g

    @classmethod
    def create(cls, **kargs):
        # if the debug is set to True globally, we don't create any
        # clones
        if get_global_debug():
            return cls(**kargs)
        gen, cached = cls.__cached_py_generator(**kargs)
        if not cached:
            return gen
        else:
            kargs["is_clone"] = True
            g = cls(**kargs)
            g.__def_instance = gen
            g.__generator.def_instance = gen.internal_generator
            return g

    # list of helper functions similar to chisel, but force good naming
    # so that we can produce a readable verilog
    def __get_port_name_type(self, port, port_type):
        if port is None:
            clock_names = self.__generator.get_ports(port_type.value)
            assert len(clock_names) > 0, str(port_type) + " signal not found"
            port = clock_names[0]
        if isinstance(port, str):
            port = self.ports[port]
        assert self.__generator.has_port(port.name)
        return port.name

    def __get_var_assert(self, var):
        if isinstance(var, str):
            var = self.vars[var]
            assert self.__generator.has_var(var.name)
        return var

    def __create_new_var(self, var_name, var_ref):
        new_var = self.var(var_name, var_ref.width, var_ref.signed,
                           var_ref.size)
        if self.debug:
            new_var.add_fn_ln(get_fn_ln())
        return new_var

    def __add_stmt_with_debug(self, block, stmt):
        if self.debug:
            stmt.add_fn_ln(get_fn_ln())
        block.add_stmt(stmt)

    def dpi(self, func_name):
        return self.__generator.dpi_function(func_name)

    def reg_next(self, var_name, var, clk=None):
        clk_name = self.__get_port_name_type(clk, PortType.Clock)
        clk = self.ports[clk_name]
        if clk_name not in self.__reg_next_stmt:
            self.__reg_next_stmt[clk_name] = self.sequential(
                (BlockEdgeType.Posedge, clk))
        var = self.__get_var_assert(var)
        new_var = self.__create_new_var(var_name, var)
        self.__add_stmt_with_debug(self.__reg_next_stmt[clk_name],
                                   new_var.assign(var))
        return new_var

    def reg_init(self, var_name, var, clk=None, reset=None, init_value=0):
        clk_name = self.__get_port_name_type(clk, PortType.Clock)
        rst_name = self.__get_port_name_type(reset, PortType.AsyncReset)
        clk = self.ports[clk_name]
        reset = self.ports[rst_name]
        if (clk_name, rst_name) not in self.__reg_init_stmt:
            seq = self.sequential((BlockEdgeType.Posedge, clk),
                                  (BlockEdgeType.Posedge, reset))
            if_stmt = seq.if_(reset)
            self.__reg_init_stmt[(clk_name, rst_name)] = if_stmt
        if_stmt = self.__reg_init_stmt[(clk_name, rst_name)]
        var = self.__get_var_assert(var)
        new_var = self.__create_new_var(var_name, var)
        self.__add_stmt_with_debug(if_stmt.else_body(), new_var.assign(var))
        self.__add_stmt_with_debug(if_stmt.then_body(),
                                   new_var.assign(init_value))
        return new_var

    def reg_enable(self, var_name, var, en, clk=None):
        clk_name = self.__get_port_name_type(clk, PortType.Clock)
        clk = self.ports[clk_name]
        if isinstance(var, str):
            var = self.ports[var]
        assert self.__generator.has_var(var.name)
        if isinstance(en, str):
            en = self.ports[en]
        assert self.__generator.has_var(en.name)
        if (clk.name, en.name) not in self.__reg_en_stmt:
            seq = self.sequential((BlockEdgeType.Posedge, clk))
            if_stmt = seq.if_(en)
            self.__reg_en_stmt[(clk_name, en.name)] = if_stmt
        if_stmt = self.__reg_en_stmt[(clk_name, en.name)]
        var = self.__get_var_assert(var)
        new_var = self.__create_new_var(var_name, var)
        self.__add_stmt_with_debug(if_stmt.then_body(), new_var.assign(var))
        return new_var

    # meta values
    def mark_stmt(self, name: str, stmt):
        if not isinstance(stmt, _kratos.StmtBlock):
            raw_stmt = stmt.stmt()
        else:
            raw_stmt = stmt
        if self.__generator.has_named_block(name):
            raise ValueError(name + " already exists")

        self.__stmt_label_mapping[name] = stmt
        self.__generator.add_named_block(name, raw_stmt)

    def get_marked_stmt(self, name):
        assert name in self.__stmt_label_mapping
        return self.__stmt_label_mapping[name]


def always_ff(*sensitivity):
    assert len(sensitivity) > 0, "always_ff needs at least one signal"
    for edge, var in sensitivity:
        assert isinstance(edge, BlockEdgeType)
        assert isinstance(var, (str, _kratos.Var))

    return AlwaysWrapper


def always_latch(fn):
    return AlwaysWrapper(fn)


def always_comb(fn):
    return AlwaysWrapper(fn)


def initial(fn):
    return AlwaysWrapper(fn)



import _kratos
from .util import const
from typing import Union, List
from _kratos import get_fn_ln


class IfStmt:
    def __init__(self, predicate: _kratos.Var):
        self._stmt = _kratos.IfStmt(predicate)
        self.__generator = predicate.generator
        if self.__generator.debug:
            self._stmt.add_fn_ln(get_fn_ln())

    def then_(self, *args: _kratos.Stmt):
        for stmt in args:
            if self.__generator.debug:
                stmt.add_fn_ln(get_fn_ln())
            if not isinstance(stmt, _kratos.Stmt):
                stmt = stmt.stmt()
            self._stmt.add_then_stmt(stmt)
        return self

    def else_(self, *args: _kratos.Stmt):
        for stmt in args:
            if self.__generator.debug:
                stmt.add_fn_ln(get_fn_ln())
            if not isinstance(stmt, _kratos.Stmt):
                stmt = stmt.stmt()
            self._stmt.add_else_stmt(stmt)
        return self

    def then_body(self):
        return self._stmt.then_body()

    def else_body(self):
        return self._stmt.else_body()

    def add_scope_variable(self, name, value, is_var=False, override=False):
        self._stmt.add_scope_variable(name, value, is_var, override)

    def stmt(self):
        return self._stmt

    def add_fn_ln(self, info):
        self._stmt.add_fn_ln(info)


def if_(predicate: _kratos.Var):
    return IfStmt(predicate)


class SwitchStmt:
    def __init__(self, predicate: _kratos.Var):
        self._stmt = _kratos.SwitchStmt(predicate)
        if predicate.generator.debug:
            self._stmt.add_fn_ln(get_fn_ln())
        self.__predicate = predicate
        self.__generator = predicate.generator

    def case_(self, cond: _kratos.Var, *args: _kratos.Stmt):
        if isinstance(cond, int):
            cond = const(cond, self.__predicate.width, self.__predicate.signed)
        case = None
        for stmt in args:
            if not isinstance(stmt, _kratos.Stmt):
                stmt = stmt.stmt()
            case = self._stmt.add_switch_case(cond, stmt)
        assert case is not None
        if self.__generator.debug:
            case.add_fn_ln(get_fn_ln())
        return self

    def stmt(self):
        return self._stmt

    def add_scope_variable(self, name, value, is_var=False, override=False):
        self._stmt.add_scope_variable(name, value, is_var, override)


def switch_(predicate: _kratos.Var):
    return SwitchStmt(predicate)


class RawStringStmt:
    def __init__(self, value: Union[str, List[str]]):
        self._stmt = _kratos.RawStringStmt(value)

    def stmt(self):
        return self._stmt



from kratos.lib import SinglePortSRAM, bank_sram
from kratos import verilog


def test_single_port_sram(check_gold):
    s = SinglePortSRAM("SRAM_MACRO", 16, 10)
    assert s.num_ports == 1
    check_gold(s, "test_single_port_sram")


def test_single_port_sram_bank(check_gold):
    s = SinglePortSRAM("SRAM_MACRO", 16, 10)
    # bank 6 of them together
    sram = bank_sram("Memory", s.capacity * 6, s)
    assert isinstance(sram, SinglePortSRAM)
    check_gold(sram, "test_single_port_sram_bank")


if __name__ == "__main__":
    from conftest import check_gold_fn
    test_single_port_sram_bank(check_gold_fn)



from _kratos import get_fn_ln


class FSM:
    """Wrapper class for _kratos.FSM"""
    def __init__(self, generator, fsm):
        self.__generator = generator
        self.__fsm = fsm
        self.__states = {}

    def output(self, var):
        self.__fsm.output(var)

    def add_state(self, name):
        if self.__generator.debug:
            debug = get_fn_ln()
            state = FSMState(self.__generator, self.__fsm.add_state(name, debug))
        else:
            state = FSMState(self.__generator, self.__fsm.add_state(name))
        self.__states[name] = state
        return state

    def set_start_state(self, state):
        if isinstance(state, FSMState):
            state = state.internal_state
        if self.__generator.debug:
            debug = get_fn_ln()
            self.__fsm.set_start_state(state, debug)
        else:
            self.__fsm.set_start_state(state)

    def dot_graph(self, filename=None):
        if filename is not None:
            return self.__fsm.dot_graph(filename)
        else:
            return self.__fsm.dot_graph()

    def output_table(self, filename=None):
        if filename is not None:
            return self.__fsm.output_table(filename)
        else:
            return self.__fsm.output_table()

    def __getitem__(self, fsm_name):
        return self.__states[fsm_name]

    def add_child_fsm(self, fsm):
        self.__fsm.add_child_fsm(fsm.__fsm)

    def realize(self):
        self.__fsm.realize()

    @property
    def current_state(self):
        return self.__fsm.current_state

    @property
    def is_moore(self):
        return self.__fsm.is_moore()

    @is_moore.setter
    def is_moore(self, value):
        self.__fsm.set_moore(value)

    def set_reset_high(self, value: bool):
        self.__fsm.set_reset_high(value)


class FSMState:
    def __init__(self, generator, fsm_state):
        self.__generator = generator
        self.__fsm_state = fsm_state

    def next(self, next_state, cond):
        if self.__generator.debug:
            debug = get_fn_ln()
            self.__fsm_state.next(next_state.internal_state, cond, debug)
        else:
            self.__fsm_state.next(next_state.internal_state, cond)

    def output(self, var, value):
        if self.__generator.debug:
            debug = get_fn_ln()
            self.__fsm_state.output(var, value, debug)
        else:
            self.__fsm_state.output(var, value)

    @property
    def internal_state(self):
        return self.__fsm_state



import _kratos
from .generator import Generator


class SinglePortSRAM(Generator):
    def __init__(self, macro_name: str, data_width: int, addr_width: int,
                 partial_write: bool = False, is_clone=False, sram_def=None):
        if sram_def is None:
            self.sram = _kratos.lib.SinglePortSRAM(Generator.get_context(),
                                                   macro_name,
                                                   addr_width, data_width,
                                                   partial_write)
        else:
            self.sram = sram_def
        Generator.__init__(self, macro_name, is_clone=is_clone,
                           internal_generator=self.sram)

    # proxy for properties
    @property
    def num_ports(self):
        return self.sram.num_ports

    @property
    def addr_width(self):
        return self.sram.addr_width

    @property
    def data_width(self):
        return self.sram.data_width

    @property
    def capacity(self):
        return self.sram.capacity()

    # ports
    # to change the name
    # simply rename the port, e.g. sram.output_data.name = "Q_DATA"
    @property
    def output_data(self):
        return self.sram.output_data

    @property
    def chip_enable(self):
        return self.sram.chip_enable

    @property
    def write_enable(self):
        return self.sram.write_enable

    @property
    def addr(self):
        return self.sram.addr

    @property
    def input_data(self):
        return self.sram.input_data

    @property
    def partial_write_mask(self):
        return self.partial_write_mask


def bank_sram(generator_name, capacity, sram_def):
    if isinstance(sram_def, SinglePortSRAM):
        sram_def = sram_def.sram
    else:
        assert isinstance(sram_def, _kratos.lib.SinglePortSRAM)
    sram = _kratos.lib.SinglePortSRAM(Generator.get_context(), generator_name,
                                      capacity, sram_def)
    # allow nested sram banks
    return SinglePortSRAM(generator_name, sram.data_width, sram.addr_width,
                          sram.partial_write, False, sram_def=sram)



from kratos import Generator, signed, unsigned, verilog
from kratos.util import reduce_or, Const, const
import _kratos.exception


def test_expr():
    mod = Generator("mod")
    a = mod.var("a", 2)
    b = mod.var("b", 2)
    expr = a + b
    assert str(expr) == "a + b"


def test_slice():
    mod = Generator("mod")
    a = mod.var("a", 2)
    b = a[0]
    assert b.width == 1
    b = a[1, 0]
    assert b.width == 2
    assert str(b) == "a[1:0]"


def test_assign():
    mod = Generator("mod")
    a = mod.var("a", 2)
    b = mod.var("b", 2)
    assign = a.assign(b)
    assert id(assign.left) == id(a)
    assert id(assign.right) == id(b)


def test_signed():
    mod = Generator("mod")
    a = mod.var("a", 2)
    c = signed(a)
    assert str(c) == "signed'(a)"


def test_unsigned():
    mod = Generator("mod")
    a = mod.var("a", 2, is_signed=True)
    c = unsigned(a)
    assert str(c) == "unsigned'(a)"


def test_reduce_or():
    mod = Generator("mod")
    a = mod.var("a", 2)
    b = mod.var("b", 2)
    expr = reduce_or(a, b)
    assert str(expr) == "a | b"
    assert str(reduce_or(a)) == "a"


def test_const():
    a = Const[4](2)
    b = const(2, 4)
    assert str(a) == str(b)
    assert str(a) == "4'h2"


def test_width():
    mod = Generator("mod", True)
    a = mod.var("a", 1)
    a.width = 2
    file_names = list(a.fn_name_ln)
    assert len(file_names) == 2
    assert file_names[0][0] == file_names[1][0]
    assert abs(file_names[0][1] - file_names[1][1]) == 1


def test_slice_same():
    mod = Generator("mod")
    a = mod.var("a", 2, size=2)
    b = a[1, 1]
    assert str(b) == "a[1]"


def test_explicit_array():
    mod = Generator("mod")
    a = mod.var("a", 2, explicit_array=True, packed=True)
    b = a[0, 0]
    assert b.width == 2
    c = mod.var("c", 2, explicit_array=True, packed=True)
    d = mod.var("d", 1)
    e = mod.var("e", 2)
    mod.wire(c, a)
    mod.wire(e, a[d])
    src = verilog(mod)["mod"]
    assert "logic [0:0][1:0] a" in src


def test_neq():
    mod = Generator("mod")
    a = mod.var("a", 2)
    b = a != 2
    assert str(b) == "a != 2'h2"


def test_nested_ternary():
    from kratos import mux
    mod = Generator("mod")
    a = mod.var("a", 1)
    b = mod.var("b", 1)
    c = a + mux(a, a, b)
    assert str(c) == "a + (a ? a: b)"


def test_packed_unpacked():
    import _kratos
    mod = Generator("mod")
    a = mod.var("a", 2, size=2, packed=True)
    b = mod.var("b", 2, size=2)
    c = const(2, 4)
    d = mod.var("d", 4, explicit_array=True)
    try:
        a.assign(b)
        assert False
    except _kratos.exception.StmtException:
        assert True
    a.assign(4)
    try:
        b.assign(c)
        assert False
    except _kratos.exception.StmtException:
        assert True
    try:
        d.assign(a)
        assert False
    except _kratos.exception.StmtException:
        assert True


def test_md_array():
    mod = Generator("mod")
    a = mod.var("a", 16, size=(4, 2), packed=True)
    b = mod.var("b", 16, size=(4, 2), packed=True)
    c = mod.var("c", 16, size=(2,), packed=True)
    mod.wire(b, a)
    mod.wire(c, a[1])
    src = verilog(mod)["mod"]
    assert "logic [3:0][1:0][15:0] a;" in src


def test_enum_cast():
    from kratos import enum
    e = enum("Type", ["a", "b"])
    mod = Generator("mod")
    var = mod.enum_var("v", e)
    casted = enum(const(1, 1), e)
    assert str(casted) == "Type'(1'h1)"
    var.assign(casted)


def test_bool_except():
    mod = Generator("mod")
    a = mod.var("a", 1)
    try:
        _ = not a
        assert False
    except _kratos.exception.InvalidConversionException:
        assert True


def test_single_concat():
    from kratos import concat
    mod = Generator("mod")
    a = mod.var("a", 1)
    r = concat([a])
    assert id(r) == id(a)


def test_resize():
    from kratos import resize
    mod = Generator("mod")
    a = mod.var("a", 1)
    b = resize(a, 16)
    assert b.width == 16
    assert str(b) == "16'(a)"


def test_power():
    mod = Generator("mod")
    a = mod.var("a", 5)
    b = 2 ** a
    assert str(b) == "5'h2 ** a"


def test_lshift():
    mod = Generator("mod")
    a = mod.var("a", 5)
    b = mod.var("b", 5)
    c = a << b
    assert str(c) == "a << b"


if __name__ == "__main__":
    test_lshift()




import os
import re
import sys
import platform
import subprocess
import multiprocessing
import shutil

from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext
from distutils.version import LooseVersion


class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=''):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)


class CMakeBuild(build_ext):
    def run(self):
        try:
            out = subprocess.check_output(['cmake', '--version'])
        except OSError:
            raise RuntimeError(
                "CMake must be installed to build the following extensions: " +
                ", ".join(e.name for e in self.extensions))

        if platform.system() == "Windows":
            cmake_version = LooseVersion(
                re.search(r'version\s*([\d.]+)', out.decode()).group(1))
            if cmake_version < '3.1.0':
                raise RuntimeError("CMake >= 3.1.0 is required on Windows")

        for ext in self.extensions:
            self.build_extension(ext)

    @staticmethod
    def is_windows():
        tag = platform.system().lower()
        return tag == "windows"

    def build_extension(self, ext):
        extdir = os.path.abspath(
            os.path.dirname(self.get_ext_fullpath(ext.name)))
        cmake_args = ['-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + extdir,
                      '-DPYTHON_EXECUTABLE=' + sys.executable]

        # test env variable to determine whether to build in debug
        if os.environ.get("KRATOS_COVERAGE") is not None or \
                os.environ.get("KRATOS_DEBUG") is not None:
            cfg = 'Debug'
        else:
            cfg = 'Release'
        build_args = ['--config', cfg]
        env = os.environ.copy()

        cmake_args += ['-DCMAKE_BUILD_TYPE=' + cfg]
        if self.is_windows():
            cmake_args += ["-G", "Unix Makefiles"]
            # make sure clang is in the PATH
            clang_path = shutil.which("clang")
            assert clang_path is not None, \
                "Unable to find clang. Currently only clang is supported "\
                "on Windows"
            cxx_path = shutil.which("clang++")
            rc_path = shutil.which("llvm-rc")
            cmake_args += ["-DCMAKE_C_COMPILER:PATH=" + clang_path]
            cmake_args += ["-DCMAKE_CXX_COMPILER:PATH=" + cxx_path]
            cmake_args += ["-DCMAKE_RC_COMPILER:PATH=" + rc_path]
            assert "MAKE_PROGRAM" in env,\
                "Currently only unix make is supported on windows"
            make_program = env["MAKE_PROGRAM"]
            cmake_args += [R"-DCMAKE_MAKE_PROGRAM=" + make_program]

        cpu_count = max(2, multiprocessing.cpu_count() // 2)
        build_args += ['--', '-j{0}'.format(cpu_count)]

        python_path = sys.executable
        cmake_args += ['-DPYTHON_EXECUTABLE:FILEPATH=' + python_path]

        env['CXXFLAGS'] = '{} -DVERSION_INFO=\\"{}\\"'.format(
            env.get('CXXFLAGS', ''),
            self.distribution.get_version())
        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)
        subprocess.check_call(['cmake', ext.sourcedir] + cmake_args,
                              cwd=self.build_temp, env=env)

        subprocess.check_call(
            ['cmake', '--build', '.', "--target", "_kratos"] + build_args,
            cwd=self.build_temp)


current_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(current_directory, 'README.rst')) as f:
    long_description = f.read()

setup(
    name='kratos',
    packages=[
        "kratos"
    ],
    version='0.0.32.4',
    author='Keyi Zhang',
    author_email='keyi@stanford.edu',
    description='Kratos is a fast hardware design language embedded in Python',
    url="https://github.com/Kuree/kratos",
    ext_modules=[CMakeExtension('_kratos')],
    cmdclass=dict(build_ext=CMakeBuild),
    zip_safe=False,
    python_requires=">=3.6",
    install_requires=[
        "astor",
    ],
    long_description=long_description,
    long_description_content_type='text/x-rst',
)



from .pyast import transform_function_block, get_fn, get_ln, \
    extract_arg_name_order_from_fn
import inspect
import _kratos


class FunctionCall:
    cache_ordering = {}

    def __init__(self, fn):
        self.__fn = fn

    def __call__(self, *args):
        # notice that we need to get self from the upper locals
        f = inspect.currentframe().f_back
        if "self" not in f.f_locals:
            raise Exception("Unable to find self from local scope")
        generator = f.f_locals["self"]
        fn_name = self.__fn.__name__
        if not generator.internal_generator.has_function(fn_name):
            # we infer the types. notice that we don't have names here
            arg_types = []
            for arg in args:
                arg_types.append((arg.width, arg.signed))
            args_order, stmts = transform_function_block(generator, self.__fn,
                                                         arg_types)

            # add statements
            func = generator.internal_generator.get_function(fn_name)
            for stmt in stmts:
                if not isinstance(stmt, _kratos.Stmt):
                    stmt = stmt.stmt()
                func.add_stmt(stmt)
            # set ordering
            func.set_port_ordering(args_order)
            self.cache_ordering[generator.name, fn_name] = args_order
            if generator.debug:
                fn, ln = get_fn(self.__fn), get_ln(self.__fn)
                for _, var_name in args_order.items():
                    port = func.get_port(var_name)
                    port.add_fn_ln((fn, ln + 1))
        else:
            args_order = self.cache_ordering[generator.name, fn_name]
        mapping = {}
        for idx, value in enumerate(args):
            var_name = args_order[idx]
            mapping[var_name] = value

        return generator.internal_generator.call(fn_name, mapping)


class DPIFunctionCall:
    cache_ordering = {}

    def __init__(self, width=0, is_pure=False, is_context=False):
        self.width = width
        if is_pure or is_context:
            assert is_pure ^ is_context, "DPI can be either pure or context"
        self.is_pure = is_pure
        self.is_context = is_context

    def __call__(self, fn):
        width = self.width
        is_pure = self.is_pure
        is_context = self.is_context

        class _DPIFunctionCall:
            def __init__(self):
                self.width = width
                self.__fn = fn

            def __call__(self, *args):
                fn_name = self.__fn.__name__
                cache = DPIFunctionCall.cache_ordering
                arg_types = []
                gen = None
                for arg in args:
                    arg_types.append((arg.width, arg.signed))
                    if gen is None:
                        gen = arg.generator
                if not gen.has_function(fn_name):
                    func = gen.dpi_function(fn_name)
                    func.set_return_width(self.width)
                    if is_pure:
                        func.set_is_pure(is_pure)
                    elif is_context:
                        func.set_is_context(is_context)
                else:
                    func = gen.get_function(fn_name)
                if fn_name not in DPIFunctionCall.cache_ordering:

                    assert gen is not None, "Unable to determine args"
                    args_order = extract_arg_name_order_from_fn(self.__fn)
                    lst = list(args_order.keys())
                    lst.sort()
                    for idx in lst:
                        name = args_order[idx]
                        w, signed = arg_types[idx]
                        func.input(name, w, signed)
                    func.set_port_ordering(args_order)
                    cache[fn_name] = args_order, gen, arg_types
                else:
                    args_order, gen, arg_types = cache[fn_name]
                mapping = {}
                for idx, value in enumerate(args):
                    var_name = args_order[idx]
                    mapping[var_name] = value

                return gen.call(fn_name, mapping)

        return _DPIFunctionCall()


def get_built_in(gen, func_name):
    class _BuiltinFunctionCall:
        def __init__(self):
            self.func_name = func_name
            self.gen = gen

        def __call__(self, *args):
            if not gen.has_function(self.func_name):
                gen.builtin_function(self.func_name)
            return gen.call(self.func_name, args)
    return _BuiltinFunctionCall()


# name alias
function = FunctionCall
dpi_function = DPIFunctionCall


def clear_context():
    FunctionCall.cache_ordering.clear()
    DPIFunctionCall.cache_ordering.clear()



from _kratos.passes import *
from .generator import Generator
import _kratos
from .debug import dump_debug_database
from typing import Dict
import os


class Attribute(_kratos.passes.Attribute):
    def __init__(self):
        _kratos.passes.Attribute.__init__(self, self)


__generated_definitions = set()


def verilog(generator: Generator, optimize_if: bool = True,
            optimize_passthrough: bool = True,
            optimize_fanout: bool = True,
            optimize_bundle: bool = True,
            reorder_stmts: bool = False,
            check_active_high: bool = True,
            debug_fn_ln: bool = False,
            additional_passes: Dict = None,
            int_dpi_interface: bool = True,
            remove_assertion: bool = False,
            check_inferred_latch: bool = True,
            check_multiple_driver: bool = True,
            check_combinational_loop: bool = True,
            insert_pipeline_stages: bool = False,
            filename: str = None,
            output_dir: str = None,
            insert_debug_info: bool = False,
            insert_verilator_info: bool = False,
            check_flip_flop_always_ff: bool = True,
            remove_unused: bool = True,
            debug_db_filename: str = "",
            ssa_transform: bool = False,
            use_parallel: bool = True,
            track_generated_definition: bool = False,
            compile_to_verilog: bool = False):
    code_gen = _kratos.VerilogModule(generator.internal_generator)
    pass_manager = code_gen.pass_manager()
    if additional_passes is not None:
        for name, fn in additional_passes.items():
            pass_manager.register_pass(name, fn)
            pass_manager.add_pass(name)
    # load all the passes
    # you can easily roll your own functions to control how the passes
    # are run
    if remove_assertion:
        pass_manager.add_pass("remove_assertion")
    pass_manager.add_pass("realize_fsm")
    if ssa_transform:
        pass_manager.add_pass("ssa_transform_fix")
    if optimize_passthrough:
        pass_manager.add_pass("remove_pass_through_modules")
    if optimize_if:
        pass_manager.add_pass("merge_if_block")
        pass_manager.add_pass("transform_if_to_case")
    # fsm elaboration has to happen before unused vars removal
    pass_manager.add_pass("zero_out_stubs")
    if optimize_fanout:
        pass_manager.add_pass("remove_fanout_one_wires")
    pass_manager.add_pass("zero_generator_inputs")
    if optimize_bundle:
        pass_manager.add_pass("change_port_bundle_struct")
    pass_manager.add_pass("verify_generator_connectivity")
    pass_manager.add_pass("decouple_generator_ports")
    pass_manager.add_pass("fix_assignment_type")
    if remove_unused and not insert_debug_info:
        pass_manager.add_pass("remove_unused_vars")
        pass_manager.add_pass("remove_unused_stmts")
    pass_manager.add_pass("verify_assignments")
    if check_combinational_loop:
        pass_manager.add_pass("check_combinational_loop")
    pass_manager.add_pass("check_mixed_assignment")
    pass_manager.add_pass("check_always_sensitivity")
    if check_inferred_latch:
        pass_manager.add_pass("check_inferred_latch")
    if check_active_high:
        pass_manager.add_pass("check_active_high")
    pass_manager.add_pass("check_function_return")
    pass_manager.add_pass("merge_wire_assignments")
    if check_multiple_driver:
        pass_manager.add_pass("check_multiple_driver")
    if check_flip_flop_always_ff:
        pass_manager.add_pass("check_flip_flop_always_ff")
    # insert debug break points if needed
    if insert_debug_info:
        pass_manager.add_pass("propagate_scope_variable")
        pass_manager.add_pass("inject_instance_ids")
        pass_manager.add_pass("inject_debug_break_points")
        pass_manager.add_pass("inject_assert_fail_exception")
    if use_parallel:
        pass_manager.add_pass("hash_generators_parallel")
    else:
        pass_manager.add_pass("hash_generators_sequential")
    pass_manager.add_pass("change_property_into_stmt")
    pass_manager.add_pass("uniquify_generators")
    pass_manager.add_pass("create_module_instantiation")
    pass_manager.add_pass("create_interface_instantiation")
    if insert_pipeline_stages:
        pass_manager.add_pass("insert_pipeline_stages")
    if reorder_stmts:
        pass_manager.add_pass("sort_stmts")

    code_gen.run_passes()

    if compile_to_verilog:
        assert output_dir is None and filename is not None,\
            "Trans-compile to verilog is only supported by a single file"
        import shutil
        assert shutil.which("sv2v") is not None,\
            "Compiling to verilog requires sv2v"

    if insert_verilator_info:
        _kratos.passes.insert_verilator_public(generator.internal_generator)

    if output_dir is not None:
        if not os.path.isdir(output_dir):
            os.makedirs(output_dir)
        package_name = generator.internal_generator.name + "_pkg"
        _kratos.passes.generate_verilog(generator.internal_generator,
                                        output_dir,
                                        package_name,
                                        debug_fn_ln)
        r = None
    else:
        src = code_gen.verilog_src()
        result = [src]
        gen = generator.internal_generator
        if debug_fn_ln:
            info = _kratos.passes.extract_debug_info(gen)
            result.append(info)
        else:
            info = {}

        # struct info
        struct_info = _kratos.passes.extract_struct_info(
            generator.internal_generator)
        if len(struct_info) > 0:
            result.append(struct_info)

        # dpi info
        dpi_func = _kratos.passes.extract_dpi_function(gen, int_dpi_interface)
        if len(dpi_func) > 0:
            result.append(dpi_func)
        enum_def = _kratos.passes.extract_enum_info(gen)
        if len(enum_def) > 0:
            result.append(enum_def)

        # interface info
        interface_info = _kratos.passes.extract_interface_info(gen)
        if len(interface_info) > 0:
            result.append(interface_info)

        if filename is not None:
            output_verilog(filename, src, info, struct_info, dpi_func, enum_def,
                           interface_info, track_generated_definition)
            if compile_to_verilog:
                pipe = os.popen("sv2v " + filename, "r")
                s = pipe.read()
                with open(filename, "w+") as f:
                    f.write(s)
            generator.internal_generator.verilog_fn = filename
        r = result[0] if len(result) == 1 else result

    # debug database
    if debug_db_filename:
        dump_debug_database(generator, debug_db_filename)

    return r


def output_verilog(filename, mod_src, info, struct_info, dpi_func, enum_def,
                   interface_info, track_generated_definition):
    line_no = 1
    debug_info = {}
    with open(filename, "w+") as f:
        definitions = [dpi_func, struct_info, enum_def, interface_info]
        for definition in definitions:
            for name, def_ in definition.items():
                if track_generated_definition \
                        and name in __generated_definitions:
                    continue
                lines = def_.split("\n")
                for line in lines:
                    f.write(line + "\n")
                    line_no += 1
                if track_generated_definition:
                    __generated_definitions.add(name)

        mod_src_names = list(mod_src.keys())
        mod_src_names.sort()
        for mod_name in mod_src_names:
            if track_generated_definition and \
                    mod_name in __generated_definitions:
                continue
            src = mod_src[mod_name]
            # get mod
            mods = Generator.get_context().get_generators_by_name(mod_name)
            for mod in mods:
                # fix the line number
                # need to subtract 1 since on the C++'s side already
                # has 1 offset
                _kratos.fix_verilog_ln(mod, line_no - 1)
            mod_info = info[mod_name] if mod_name in info else {}
            lines = src.split("\n")
            for index, line in enumerate(lines):
                f.write(line + "\n")
                if index + 1 in mod_info:
                    debug_info[line_no] = mod_info[index + 1]
                line_no += 1
            if track_generated_definition:
                __generated_definitions.add(mod_name)

    if len(debug_info) > 0:
        with open(filename + ".debug", "w+") as f:
            import json
            json.dump(debug_info, f)


def clear_context():
    Generator.clear_context()
    __generated_definitions.clear()



from .generator import Generator, PortType, PortDirection, BlockEdgeType, \
    initial, CombinationalCodeBlock, SequentialCodeBlock, \
    set_global_debug, enum, always_comb, always_ff, always_latch, has_enum

from .passes import Attribute, verilog, clear_context
from .util import clog2, reduce_add, reduce_and, reduce_mul, reduce_or, \
    concat, ext, mux, ternary, signed, const, comment, unsigned, create_stub, \
    resize

from .ports import PortBundle
from .tb import TestBench, assert_, delay, assume, cover
from .debug import enable_runtime_debug
from .pyast import add_scope_context
from .sim import Simulator

# directly import from the underlying C++ bindingtest_generator_proper
from _kratos.util import is_valid_verilog
from _kratos.exception import VarException, StmtException
from _kratos.passes import IRVisitor as IRVisitor
from _kratos import PackedStruct, Port, Var, AssignmentType, VarSlice, \
    VarVarSlice
from _kratos import Sequence
from _kratos import DebugDataBase
from .stmts import IfStmt, SwitchStmt, if_, switch_, RawStringStmt
from _kratos import Interface
from _kratos import PropertyAction, ParamType

# FSMs
from _kratos import FSM, FSMState

# fault
# TODO fix this namespace
from _kratos.fault import SimulationRun, FaultAnalyzer,\
    parse_verilator_coverage, parse_icc_coverage

posedge = BlockEdgeType.Posedge
negedge = BlockEdgeType.Negedge

__all__ = ["Generator", "PortType", "PortDirection", "BlockEdgeType",
           "verilog", "const", "is_valid_verilog", "VarException",
           "StmtException", "IRVisitor", "FSM", "FSMState", "initial",
           "Sequence", "TestBench", "assert_", "delay", "enable_runtime_debug",
           "enum", "clear_context", "always_comb", "always_ff", "always_latch",
           "has_enum"]

__all__ += ["CombinationalCodeBlock", "SequentialCodeBlock", "SwitchStmt",
            "PackedStruct", "Port", "Var", "IfStmt", "AssignmentType",
            "if_", "switch_", "Attribute", "PortBundle", "DebugDataBase",
            "add_scope_context", "set_global_debug", "Interface", "VarSlice",
            "VarVarSlice", "ParamType"]

# utils
__all__ += ["clog2", "reduce_add", "reduce_and", "reduce_mul", "reduce_or",
            "concat", "ext", "comment", "signed", "unsigned", "create_stub",
            "resize"]

# type aliasing
__all__ += ["BlockEdgeType", "posedge", "negedge", "mux", "ternary"]
